// ABBYY ® Mobile Imaging UI Components © 2018 ABBYY Production LLC.
// ABBYY is either a registered trademark or a trademark of ABBYY Software Ltd.

package com.abbyy.mobile.sample.imagecapture;

import android.content.DialogInterface;
import android.content.Intent;
import android.graphics.Bitmap;
import android.os.Bundle;
import android.support.annotation.NonNull;
import android.support.v7.app.AlertDialog;
import android.support.v7.app.AppCompatActivity;
import android.view.View;
import android.widget.ImageButton;
import android.widget.Toast;

import com.abbyy.mobile.rtr.Engine;
import com.abbyy.mobile.sample.imagecapture.image.ImageUtils;
import com.abbyy.mobile.sample.imagecapture.multipage.CaptureResultDialogFragment;
import com.abbyy.mobile.sample.imagecapture.multipage.MultiPageCounter;
import com.abbyy.mobile.uicomponents.CaptureView;
import com.abbyy.mobile.uicomponents.scenario.ImageCaptureScenario;

/**
 * This activity uses UI component to capture a page, shows a dialog to confirm or discard
 * the result, and keeps count
 */
public class CaptureActivity extends AppCompatActivity implements ImageCaptureScenario.Callback,
	CaptureResultDialogFragment.ResultListener {

	public static final String RESULT_SHOWN_EXTRA = "result_shown";

	//Finish capture view and its children
	private MultiPageCounter multiPageCounter;

	// Capture view component
	private CaptureView captureView;
	private ImageCaptureScenario imageCaptureScenario;

	// This dialog is shown when user wants to leave without saving
	private AlertDialog discardPagesDialog;

	// Whether the result dialog is currently shown
	private boolean resultDialogShown = false;

	private int alreadySavedPageCount;

	@Override
	protected void onCreate( Bundle savedInstanceState )
	{
		super.onCreate( savedInstanceState );

		if( savedInstanceState != null ) {
			resultDialogShown = savedInstanceState.getBoolean( RESULT_SHOWN_EXTRA, false );
		}

		setContentView( R.layout.activity_capture );

		// The 'ABBYY MI SDK Engine' to be used in this sample application
		Engine engine = EngineFactory.createEngine( this );

		initCaptureView( engine );

		// Initialize page counter view
		initPageCounterView();

		ImageButton cancelCapture = findViewById( R.id.cancelCapture );
		cancelCapture.setOnClickListener( new View.OnClickListener() {
			@Override
			public void onClick( View v )
			{
				finishWithWarning();
			}
		} );

		alreadySavedPageCount = ImageUtils.getPageCount( this );
	}

	@Override protected void onDestroy()
	{
		multiPageCounter.cancelLoad();
		super.onDestroy();
	}

	// Supporting correct behavior on activity re-creation
	@Override
	protected void onSaveInstanceState( Bundle outState )
	{
		outState.putBoolean( RESULT_SHOWN_EXTRA, resultDialogShown );
		super.onSaveInstanceState( outState );
	}

	private void initPageCounterView()
	{
		multiPageCounter = findViewById( R.id.finishCapture );
		multiPageCounter.setOnClickListener( new View.OnClickListener() {
			@Override
			public void onClick( View v )
			{
				finishCapture();
			}
		} );

		final int sessionCapturePageCount = ImageUtils.getCaptureSessionPageCount( this );
		multiPageCounter.updateUI( sessionCapturePageCount, null );
	}

	private void initCaptureView( Engine engine )
	{
		captureView = findViewById( R.id.captureView );
		if( engine != null ) {
			imageCaptureScenario = new ImageCaptureScenario( engine );
			imageCaptureScenario.setCallback( this );

			captureView.setCaptureScenario( imageCaptureScenario );

			startCapture();
		}
	}

	private void startCapture()
	{
		if( !resultDialogShown ) {
			imageCaptureScenario.start();
		}
	}

	private void stopCapture()
	{
		if( !resultDialogShown ) {
			imageCaptureScenario.stop();
		}
	}

	@Override
	protected void onResume()
	{
		super.onResume();
		captureView.startCamera();
	}

	@Override
	protected void onPause()
	{
		captureView.stopCamera();
		if( discardPagesDialog != null ) {
			discardPagesDialog.dismiss();
		}
		super.onPause();
	}

	// This is a callback from UI component on successful image capture
	@Override
	public void onImageCaptured( @NonNull ImageCaptureScenario.Result documentCaptureResult )
	{
		// New page number is equal to the capture session page count
		int newPageNumber = ImageUtils.getCaptureSessionPageCount( this );
		CaptureResultDialogFragment resultFragment = CaptureResultDialogFragment.newInstance(
			newPageNumber, alreadySavedPageCount, documentCaptureResult.getBitmap()
		);
		resultFragment.show( getSupportFragmentManager(), "result" );

		resultDialogShown = true;
	}

	// Error handler for UI Component
	@Override
	public void onError( @NonNull Exception error )
	{
		String errorMessage;
		if( error.getMessage() != null ) {
			errorMessage = error.getMessage();
		} else {
			errorMessage = getString( R.string.unknown_error );
		}
		Toast.makeText( this, errorMessage, Toast.LENGTH_SHORT ).show();
	}

	// Callback methods for CaptureResultDialogFragment

	@Override
	public void onCaptureResultConfirmed( Bitmap capturedPage, boolean confirmed, boolean shouldFinishCapture )
	{
		resultDialogShown = false;
		if( confirmed ) {
			int sessionCapturePageCount = ImageUtils.getCaptureSessionPageCount( this );
			multiPageCounter.updateUI( sessionCapturePageCount, capturedPage );

			if( shouldFinishCapture ) { // Finish capture session & return to the main activity
				finishCapture();
				return;
			}
		}
		startCapture(); // Resume capture
	}

	private void finishWithWarning()
	{
		int addPageCount = ImageUtils.getCaptureSessionPageCount( this );
		if( addPageCount > 0 ) {
			// There are pages that would be discarded, needs confirmation
			showDiscardPagesDialog();
		} else {
			setResult( RESULT_CANCELED, new Intent() );
			finish();
		}
	}

	private void showDiscardPagesDialog()
	{
		stopCapture(); // Pause capture until dialog is dismissed
		AlertDialog.Builder builder = new AlertDialog.Builder( this, R.style.AlertDialog );
		builder.setMessage( R.string.captured_pages_delete_warning )
			.setTitle( R.string.discard_pages )
			.setPositiveButton( R.string.discard, new DialogInterface.OnClickListener() {
				@Override
				public void onClick( DialogInterface dialog, int which )
				{
					ImageUtils.clearCaptureSessionPages( CaptureActivity.this );
					setResult( RESULT_CANCELED, null );
					finish();
				}
			} )
			.setNegativeButton( R.string.cancel, null )
			.setOnDismissListener( new DialogInterface.OnDismissListener() {
				@Override
				public void onDismiss( DialogInterface dialog )
				{
					startCapture();
				}
			} );
		discardPagesDialog = builder.show();
	}

	@Override
	public void onBackPressed()
	{
		finishWithWarning();
	}

	private void finishCapture()
	{
		ImageUtils.renameCaptureSessionPages( this );
		setResult( RESULT_OK, new Intent() );
		finish();
	}
}
