// ABBYY ® Mobile Imaging UI Components © 2018 ABBYY Production LLC.
// ABBYY is either a registered trademark or a trademark of ABBYY Software Ltd.

package com.abbyy.mobile.sample.imagecapture.image;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Matrix;
import android.media.ThumbnailUtils;
import android.support.v4.graphics.drawable.RoundedBitmapDrawable;
import android.support.v4.graphics.drawable.RoundedBitmapDrawableFactory;
import android.util.DisplayMetrics;

import com.abbyy.mobile.sample.imagecapture.R;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Comparator;

/**
 * Utility class for handling & storing images
 */
public class ImageUtils {

	// Synchronously save bitmap to file
	public static void saveBitmap( Bitmap image, File file ) throws IOException
	{
		try( FileOutputStream fos = new FileOutputStream( file ) ) {
			image.compress( Bitmap.CompressFormat.JPEG, 100, fos );
			fos.flush();
		}
	}

	// Load bitmap and scale it to fit required size
	public static Bitmap loadThumbnail( File file, int requiredDimension ) throws IOException
	{
		int ratio = getThumbnailRatio( file, requiredDimension );

		BitmapFactory.Options options = new BitmapFactory.Options();
		options.inSampleSize = ratio;
		options.inJustDecodeBounds = false;

		Bitmap image;
		try( FileInputStream fis = new FileInputStream( file ) ) {
			image = BitmapFactory.decodeStream( fis, null, options );
		}
		return ThumbnailUtils.extractThumbnail( image, requiredDimension, requiredDimension );
	}

	private static int getThumbnailRatio( File file, int requiredDimension ) throws IOException
	{
		int ratio = 1;
		BitmapFactory.Options options = new BitmapFactory.Options();
		options.inJustDecodeBounds = true;
		try( FileInputStream fis = new FileInputStream( file ) ) {
			BitmapFactory.decodeStream( fis, null, options );
		}
		int height = options.outHeight;
		int width = options.outWidth;

		if( height > requiredDimension || width > requiredDimension ) {
			int heightRatio = Math.round( (float) height / (float) requiredDimension );
			int widthRatio = Math.round( (float) width / (float) requiredDimension );

			ratio = Math.min( heightRatio, widthRatio );
		}
		return ratio;
	}

	public static File getPageFile( int pageIndex, Context context )
	{
		File pagesDir = getPagesDir( context );
		if( !pagesDir.exists() ) {
			pagesDir.mkdir();
		}
		return new File( pagesDir, "page_" + ( pageIndex + 1 ) + ".jpg" );
	}

	private static File getPagesDir( Context context )
	{
		return new File( context.getFilesDir(), "pages" );
	}

	public static File getCaptureSessionPageFile( int pageIndex, Context context )
	{
		File captureSessionDir = getCaptureSessionDir( context );
		if( !captureSessionDir.exists() ) {
			captureSessionDir.mkdir();
		}
		return new File( captureSessionDir, "page_" + ( pageIndex + 1 ) + ".jpg" );
	}

	private static File[] getPages( Context context )
	{
		File pagesDir = getPagesDir( context );
		return pagesDir.exists() ? pagesDir.listFiles() : new File[] {};
	}

	private static File[] getCaptureSessionPages( Context context )
	{
		File captureSessionDir = getCaptureSessionDir( context );
		return captureSessionDir.exists() ? captureSessionDir.listFiles() : new File[] {};
	}

	private static File getCaptureSessionDir( Context context )
	{
		return new File( context.getFilesDir(), "session_pages" );
	}

	public static void renameCaptureSessionPages( Context context )
	{
		File[] captureSessionPages = getCaptureSessionPages( context );
		// Sort by last modified date to keep order
		Arrays.sort( captureSessionPages, new Comparator<File>() {
			@Override
			public int compare( File o1, File o2 )
			{
				return Long.compare( o1.lastModified(), o2.lastModified() );
			}
		} );
		int pageCount = getPageCount( context );
		for( File captureSessionPage : captureSessionPages ) {
			File page = getPageFile( pageCount, context );
			captureSessionPage.renameTo( page );
			pageCount++;
		}
	}

	public static int getPageCount( Context context )
	{
		return getPages( context ).length;
	}

	public static int getCaptureSessionPageCount( Context context )
	{
		return getCaptureSessionPages( context ).length;
	}

	public static void clearPages( Context context )
	{
		clearCaptureSessionPages( context );
		for( File page : getPages( context ) ) {
			page.delete();
		}
	}

	public static void clearCaptureSessionPages( Context context )
	{
		for( File captureSessionPage : getCaptureSessionPages( context ) ) {
			captureSessionPage.delete();
		}
	}

	// Calculates thumbnail width depending on the number of items in a row
	public static int calculateThumbnailWidth( Context context )
	{
		DisplayMetrics displayMetrics = context.getResources().getDisplayMetrics();
		int columnsCount = context.getResources().getInteger( R.integer.columns_count );
		float screenWidth = displayMetrics.widthPixels;
		return (int) ( screenWidth / columnsCount );
	}

	// Constructs drawable with rounded corners from bitmap
	public static RoundedBitmapDrawable getRoundedImage( Context context, Bitmap originalImage )
	{
		if( originalImage == null ) {
			return null;
		}
		RoundedBitmapDrawable roundedBitmapDrawable = RoundedBitmapDrawableFactory.create( context.getResources(), originalImage );
		roundedBitmapDrawable.setCornerRadius( context.getResources().getDimension( R.dimen.corners_radius ) );
		return roundedBitmapDrawable;
	}

	public static Bitmap getMiniature( Bitmap page, int miniatureSize )
	{
		int width = page.getWidth();
		int height = page.getHeight();
		float scaleWidth = ( (float) miniatureSize ) / width;
		float scaleHeight = ( (float) miniatureSize ) / height;

		Matrix matrix = new Matrix();
		matrix.postScale( scaleWidth, scaleHeight );

		return Bitmap.createBitmap( page, 0, 0, width, height, matrix, false );
	}
}
