// ABBYY ® Mobile Imaging UI Components © 2018 ABBYY Production LLC.
// ABBYY is either a registered trademark or a trademark of ABBYY Software Ltd.

package com.abbyy.mobile.sample.imagecapture.multipage;

import android.graphics.Bitmap;
import android.os.AsyncTask;
import android.support.annotation.NonNull;
import android.support.v7.widget.RecyclerView;
import android.util.LruCache;
import android.view.ViewGroup;
import android.widget.FrameLayout;
import android.widget.ImageView;

import com.abbyy.mobile.sample.imagecapture.R;
import com.abbyy.mobile.sample.imagecapture.image.ImageLoader;
import com.abbyy.mobile.sample.imagecapture.image.ImageUtils;

import java.io.File;
import java.io.IOException;
import java.lang.ref.WeakReference;

/**
 * Adapter for pages' recycler view, with memory cache and background loading of page thumbnails
 */
public class PagesAdapter extends RecyclerView.Adapter<PagesAdapter.ViewHolder> {
	private int pagesCount;

	private static class MemoryCappedCache extends LruCache<Integer, Bitmap> {
		public MemoryCappedCache( int maxSize )
		{
			super( maxSize );
		}

		protected int sizeOf( Integer key, Bitmap value )
		{
			return value.getByteCount();
		}
	}

	// Cache for page thumbnails
	private MemoryCappedCache thumbnailCache;

	public PagesAdapter( int pagesCount )
	{
		this.pagesCount = pagesCount;
		int maxMemory = (int) ( Runtime.getRuntime().maxMemory() );

		// Use 1/8th of the available memory for the memory cache.
		int cacheSize = maxMemory / 8;
		thumbnailCache = new MemoryCappedCache( cacheSize );
	}

	public void releasePages()
	{
		thumbnailCache.evictAll();
		pagesCount = 0;
		notifyDataSetChanged();
	}

	public void updatePagesCount( int pagesCount )
	{
		if( this.pagesCount != pagesCount ) {
			this.pagesCount = pagesCount;
			notifyDataSetChanged();
		}
	}

	@NonNull
	@Override
	public ViewHolder onCreateViewHolder( @NonNull ViewGroup parent, int viewType )
	{
		// Constructing thumbnail view
		FrameLayout container = new FrameLayout( parent.getContext() );
		ImageView thumbnailView = new ImageView( parent.getContext() );
		int thumbnailWidth = ImageUtils.calculateThumbnailWidth( parent.getContext() );
		int padding = parent.getResources().getDimensionPixelOffset( R.dimen.thumbnail_padding );
		thumbnailView.setPadding( padding, padding, padding, padding );
		thumbnailView.setBackgroundResource( R.drawable.thumbnail_outline );
		thumbnailView.setCropToPadding( true );
		thumbnailView.setScaleType( ImageView.ScaleType.FIT_XY );

		int margin = parent.getResources().getDimensionPixelOffset( R.dimen.thumbnail_margin );
		FrameLayout.LayoutParams layoutParams = new FrameLayout.LayoutParams(
			ViewGroup.LayoutParams.WRAP_CONTENT, ViewGroup.LayoutParams.WRAP_CONTENT
		);
		layoutParams.setMargins( margin, margin, margin, margin );
		container.addView( thumbnailView, layoutParams );

		return new ViewHolder( container, thumbnailView, thumbnailCache, thumbnailWidth );
	}

	@Override
	public void onBindViewHolder( @NonNull final ViewHolder viewHolder, int position )
	{
		viewHolder.bind( position );
	}

	@Override
	public int getItemCount()
	{
		return pagesCount;
	}

	// Holder for thumbnail view
	static class ViewHolder extends RecyclerView.ViewHolder {
		private ImageView thumbnail;
		// Asynchronous loader
		private ImageLoader loader;
		// Bitmap cache held by weak reference
		private WeakReference<MemoryCappedCache> bitmapCacheRef;
		// Thumbnail is square, so one dimension is stored
		private int thumbnailDimension;
		private ImageLoader.Callback callback = new ImageLoader.Callback() {
			@Override
			public Bitmap loadBitmap( @NonNull File file ) throws IOException
			{
				return ImageUtils.loadThumbnail( file, thumbnailDimension );
			}

			@Override
			public void onImageReady( @NonNull Bitmap thumbnail )
			{
				// Thumbnail loaded, update view and store in cache
				updateAndNotify( thumbnail );
			}

			@Override
			public void onError( @NonNull Exception error )
			{
				// Thumbnail not loaded, update view to empty
				updateAndNotify( null );
			}
		};

		ViewHolder( @NonNull FrameLayout container, ImageView thumbnailView, MemoryCappedCache bitmapCache, int thumbnailDimension )
		{
			super( container );
			thumbnail = thumbnailView;
			this.bitmapCacheRef = new WeakReference<>( bitmapCache );
			this.thumbnailDimension = thumbnailDimension;
		}

		public void bind( int position )
		{
			MemoryCappedCache bitmapCache = bitmapCacheRef.get();
			if( bitmapCache == null ) {
				return;
			}
			Bitmap bitmap = bitmapCache.get( position );
			if( bitmap == null ) {
				if( loader != null ) {
					loader.cancel( true );
				}
				loader = getLoader( position );
				loader.executeOnExecutor( AsyncTask.THREAD_POOL_EXECUTOR );
			} else {
				thumbnail.setImageBitmap( bitmap );
			}
		}

		private ImageLoader getLoader( int position )
		{
			String pageFilePath = ImageUtils.getPageFile( position, thumbnail.getContext() ).getPath();
			return new ImageLoader( pageFilePath, callback );
		}

		private void updateAndNotify( Bitmap image )
		{
			thumbnail.setImageBitmap( image );
			MemoryCappedCache bitmapCache = bitmapCacheRef.get();
			if( bitmapCache != null && image != null ) {
				bitmapCache.put( getAdapterPosition(), image );
			}
		}
	}
}
